#!/usr/bin/perl

=head1 Name

iptc.pl -- interface for reading or adding IPTC data to image files

Reads/Adds/Updates IPTC header data to/from files (presumably image files).
    Full details are in the $help string below. (Or type "iptc -help")

Author: Dan Heller <argv@danheller.com>
	http://www.danheller.com/
	Version: 3.0 (March 29, 2004)

NOTE:
    Requires IPTCInfo.pm written by Josh Carter <josh@multipart-mixed.com>
    You can get this package on http://www.cpan.org.

=cut

$usage = qq(

iptc.pl [options] img [img2 ... ] [ directory1 directory2 ...]

    Reads or writes IPTC data to files, filling in fields in the ITPC
    structure if fields are specified in commandline params.

options:
    -help		print very verbose help
    -usage		just print the command line options (this list)
    -save		save file with changes (default: just report, no save)
    -dir		directory to save modified img (default: same file)
    -autokeys		automatically build keywords
    -autocap		if img has a .txt sister file, use contents as caption
    -caption "string"	set the "caption" to "string"
    -capfile filename	set the caption from the text in filename
    -country country	Add "country" Location to IPTC header
    -city city		Add "city" Location to IPTC header
    -format str		Output in format described in help section
    -state state	Add "state" Location to IPTC header
    -keyword "word"	Add "keyword" (to add more than one word, see below)
    -keyword "-word"	Remove "keyword" (keyword is removed if starts with -)
    -keysep "str"	String (character) to separate keywords for report_fmt
    -noprune		Don't prune keys in images/Keywords (stdout only)
    -newkeys		do not retain pre-existing keywords
    -recurse		Recursively search directories, searching for .jpg/.gif
    -reset		Clear all fields (set new values if given)
    -keyfile file	Add keywords found in "file" (default file: "Keywords")
    -keytrans file	Keyword Translation file (see below)
    -stop dir		Stop upward parse of $keyfile's when "dir" reached
    -byline "name"	Add "name" as the by-line
    -credit "string"	Add "string" as the credit
    -copyright "string"	Add "string" as the "copyright notice"
    -verbose		Report added keywords and other activity
);

$help = qq(
    The purpose of this script is to read, export, or write IPTC data
    associated with images from images, into images, or onto external
    data storage files. To understand WHY these options are important,
    see the final notes at the end of all this help.

    By default, this script examines IPTC data in images and writes it
    to stdout. Example:

    $ iptc.pl image1.jpg

    This will print all available information stored in the iamge's IPTC
    header to STDOUT.  You can modify the format of the output, as well as
    which fields you'd like to see using the -format option described later.

    You can also add or modify fields in the IPTC header.
    For example, to add byline and copyright info to an image:

    $ iptc.pl -byline "Dan Heller" -copyright "(c) www.danheller.com" img.jpg

    Note: this does not actually save the image with this new info. You
    must use -save for that, which is discussed later.

Adding Keywords to images:
    One of the main reasons for this script is to address the incredibly
    difficult, cumbersome and manually laborious process of "keywording"
    your images. However, with the aid of good image naming schemes and
    filesystem directory-name schemes, a great deal of work can be done
    for you, which this script helps you do.

    The simplest way to add a keyword to an image is to embed it directly
    into the IPTC header using the command-line option "-keyword". Here,
    the next token is used as word or words to add. If you want to add
    multiple words, put the entire list in quotes and use commas to
    separate each word. E.g.,:

    $iptc.pl -keyword "hoof, horse, animal" horse.jpg

    Assuming keywords already exist in the image that shouldn't be there,
    you can delete a word using a "-" in front of the word. E.g.,:

    $iptc.pl -keyword "horse, -turtle, animal" horse.jpg

    Here, "-turle" is removed.

    As you can see, it can be quite laborious to do this by hand for
    each image, especially when you have batches of images that would
    take (for the most part) most of the same keywords. Say you had
    several hundred images of horses.

    Automatic keywording is done by adding tree hierarchy path elements,
    and by searching for files in the tree hierarchy as described below.

    The file defined by -keyfile contains a list of keywords separated
    by commas or newlines is opened and each word added to the keywords
    for the image. -keyfile defaults to "Keywords".

    If -keyfile starts with a "/" it is a full pathname, and only that
    file is read. If it does not start with a "/", then the file is
    considered a local file, and the script will look for a file by that
    name in each directory node in the tree starting from where the image
    resides, upwards till the directory defined by -stop is reached.

    For example, if you run the command on images/Weddings/TheSmiths/*.jpg
    then the file named "Keywords" is examined in "TheSmiths", "Weddings"
    and "images." So, the example above can have a file called "Keywords"
    in the direoctry TheSmiths that has:

	joan, john, melford county, june, summer

    In the "Weddings" directory, the Keyword file could contain the words:
	special day, smiles, happy couple, it'll never last

    In the images directory, the Keyword file will apply to all images
    throughout the entire database, so be careful what you put in there,
    since no keyword in that location will be unique to any subset of
    images. That is, ALL images will contain ALL keywords in the Keywords
    file that resides in "images." More on this later.

    You can turn off this feature by using a null value: -keyfile ""

    The entire path's node names, including the target, are added as keywords.
    Example:
	images/UnitedStates/California/SanFrancisco/CableCars/trolly-1.jpg
    Keywords added:
	images, united states, california, san francisco, cable cars, trolly

    COMPLETE Directory path elements will be added, regardless of -stop.
    (-stop only applies to looking for -keyfile files.)

    When converting to keywords, these rules apply to paths AND filenames:
	Underscores (_) are converted to spaces.
	One- and Two-letter words, and any ending digits are pruned.
	Keywords are always converted to lower case.
	Words that have multiple capital letters are split into separate words,
	but each are in the _same_ keyword (see "SanFrancisco" example above).

-newkeys
    Newkeys ignores all keys currently inside the IPTC header. This flag
    is usually done because you're regenerating keys using -autokeys, or
    with -reset, where you want to clear the entire header.

Translating Keywords (-keytrans)
    To help control which keywords are used, to block certain words, or
    to convert one word to another (such as abbreviations), specify a
    keyword translation using: -keytrans filename

    The default filename is "KeywordTranslations", and contains lists of
    translations in this format:

    # this is a comment--lines starting with # are ignored, so are blank lines
    female => girl
    girl => woman
    woman => women
    # tokens can be recursive, so "female" eventually goes to "women"
    ppl => people
    cat => cats
    mama => mother
    dada => father
    # words to be deleted have no right-hand token
    the =>
    this =>
    that =>

    Images that end in "big.jpg" will NOT get the word "big" added.
    (I name my hi-res files with "big" extensions, and low-res with "sm.jpg")
    Add the word "big" to Keywords file if you want it, or make sure it's
    part of the image name without the "-" extension. Ie., "BigMoma-big.jpg"

    Keywords are sorted alphabetically and duplicates removed.

Captioning Images (-caption, -autocap)
    You can add a caption for an image using -caption "long text here...."
    If you're processing batches of files at once, you can provide individual
    captions for images by putting that text in a file with the following
    naming convention:
	img.jpg		# original image
	img.txt		# caption file for img.jpg

    Note that -autocap ALWAYS discards image's existing IPTC caption text.
    If a ".txt" file doesn't exist for an image, and you're batch processing
    many at once using -autocap on the command line, all images without a
    .txt file will get their captions erased. The lesson here is that .txt
    is inteded to be used when you typically Batch-process images, and not
    when you're editing individual captions directly into the image itself.

Output (-format -dir)
    To simply examine the iptc data for an image,
	iptc.pl image.jpg
    All non-null values are printed.  If you just want to see a preformatted
    "terse" listing for images whose iptc fields are heavily populated, use
    -terse. This only lists the file's name, path, location, by-line (author)
    and keywords.

    You can reformat the output and selecte which fields you want printed
    using -format.

    Example:
	iptc.pl -format "_path_:_object name_:_location_:_keywords_" *.jpg

    Here, each _token_ is replaced by the itpc field equivalent, with the
    following convenience tokens made for you:
	_location_ is replaced by the string, "city, state, country"
	_state_ is replaced by 'province/state'
	_country_ is replaced by 'country/primary location name'
	_path_ is replaced by the pathname to the image as typed on the cmdline
	_fullpath_ is replaced by the full pathname to the image

    All other attributes are taken directly from the IPTC values.
    If a value is not present, a null string is replaced.
    "\\t" is replaced by a TAB char, and "\\n" is replaced by a newline

    Normally, stdout will list keywords (if requested) but will prune
    out those listed in "images/Keywords", the top of the image tree.
    The reason for this is because if you are compiling a huge file
    of all the keyword data for all your images, where you can later
    search that file for images that contain certain keywords, it doesn't
    make sense to include keywords that will apply to ALL images. This
    may not apply to you if you don't rigorously apply keywords to all
    images like this. But, for those who do, -noprune can be used to
    avoid this pruning.  ALL keywords will be listed in STDOUT.

    -noprune does NOT affect the keywords that are saved with the image using
    -save discussed in the next section.

Updating Images (-save, -dir)
    Regardless of anything you do, no information is written back to the
    image's IPTC header itself unless you use -save. You can use -dir to
    specify a NEW location for the updated image, leaving the original
    unchanged. Only when -save is used *without* -dir is the original file
    overwritten with the new iptc data. It is advisable to experiment with
    using -dir and examine the resulting images before you commit to
    changing your original images, ESPECIALLY if you've manually edited
    information using a more laborious method, like Photoshop.

Final Notes:
    The struggle with deciding on the best way to manage image data for
    massive amounts of images is deciding whether to hold information about
    images *ouside* of the image itself, or to embed the data *within* the
    image. There are advantages and disadvantages to both, and once anyone
    starts down this road, they realize it's a quandry. The intent of IPTC
    is to store that info in the image, and the intent of a database is to
    store the info elsewhere. However you view the subject, the purpose of
    THIS perl code is to help you take advantage of either approach. The
    danger of using this code is because you can accidentally clobber
    precious data if you try to employe BOTH approaches, or you inadvertently
    intermix them, thinking that you're being clever. Believe me, there's
    nothing worse than realizing you've wasted days or weeks of manual labor
    by keywording your images, only to accidentally overwrite it by batch-
    processing hundreds or thousands of iamges unintentionally. Therefore,
    the best of all worlds is to always store at least a *copy* of all data
    outside image files, even if you choose to incorporate data embedded into
    images directly. If you can at least restore data, that's best. Again,
    that's where this script can be very useful.
);

use Image::IPTCInfo;
use File::Find;
use File::Basename;
use Cwd;
use Getopt::Long;

my $Byline = "";
my $Credit = "";
my $Copyright = "";
my $Country = "";
my $City = "";
my $State = "";
my $Caption = "";

my $noprune = 0;	# Prune or NOT the toplevel keys in stdout
my @masterkeys;		# keywords assigned once, reset @keywords to this
my @keywords;		# all keywords to be associated with an image
my $rm_words = {};	# remove these keywords
my %topkeys = {};	# keywords found at "$stop/Keywords" (top of tree)
my @arg_keywords = ();	# keywrods given on the commandline
my $arg_keys = "";	# in string form
my $capfile = "";
my $autocap = 0;
my $autokeys = 0;	# automatically generate keywords (see help)
my $lastpath = "";	# to avoid redundantly reading path's keywords

my $keyword_sep = ", ";	# string to separate keywords for report_fmt
my $report_fmt = "";
my $terse = 0;
my $terse_fmt = "Object name: _object name_\nPath: _path_\nLocation: _location_\nCaption: _caption/abstract_\nKeywords: _keywords_\n\n";
			# Format string to configure output (prep for DB)
my $save = 0;
my $dir = "";
my $stop = "\\.";
my $key_trans = "KeywordTranslations";	# keyword translation file
my $keyfile = "Keywords";
my $reset = 0; # reset (overwrite) iptc data (default: preserve existing data)
my $newkeys = 0; # reset keywords (default: preserve existing words)
my $recurse = 0;
my $bail = 0;		# if command line args are bad
my $verbose = 0;
my $do_usage = 0;

my @iptc_headers = (
    "object name",               "originating program",
    "edit status",               "program version",
    "editorial update",          "object cycle",
    "urgency",                   "by-line",
    "subject reference",         "by-line title",
    "category",                  "city",
    "fixture identifier",        "sub-location",
    "content location code",     "province/state",
    "content location name",     "country/primary location code",
    "release date",              "country/primary location name",
    "release time",              "original transmission reference",
    "expiration date",           "headline",
    "expiration time",           "credit",
    "special instructions",      "source",
    "action advised",            "copyright notice",
    "reference service",         "contact",
    "reference date",            "caption/abstract",
    "reference number",          "writer/editor",
    "date created",              "image type",
    "time created",              "image orientation",
    "digital creation date",     "language identifier",
    "digital creation time"
);

GetOptions(
    "help"	=>	\$bail,
    "autocap"	=>	\$autocap,
    "autokeys"	=>	\$autokeys,
    "byline=s"	=>	\$Byline,
    "format=s"	=>	\$report_fmt,
    "credit=s"	=>	\$Credit,
    "copyright=s"=>	\$Copyright,
    "caption=s"	=>	\$Caption,
    "capfile=s"	=>	\$capfile,
    "save"	=>	\$save,
    "stop=s"	=>	\$stop,
    "dir=s"	=>	\$dir,
    "country=s"	=>	\$Country,
    "city=s"	=>	\$City,
    "reset"	=>	\$reset,
    "newkeys"	=>	\$newkeys,
    "noprune"	=>	\$noprune,
    "recurse"	=>	\$recurse,
    "keytrans=s"=>	\$key_trans,
    "keyword=s"	=>	\$arg_keys,
    "keyfile=s"	=>	\$keyfile,
    "keysep=s"	=>	\$keyword_sep,
    "state=s"	=>	\$State,
    "terse"	=>	\$terse,
    "province=s"=>	\$State,
    "verbose"	=>	\$verbose,
    "usage"	=>	\$do_usage,
    ) or $bail = 2;

if ($bail) {
    if ($bail == 1) { print "$usage$help" }
    else { print "$0: Bad Parameter. Use --help for help.\n" }
    exit 1;
}

if ($do_usage) {
    print $usage;
    exit 0;
}

if ($Caption && $capfile) {
    print "$0: specify only one of -caption or -capfile\n";
    exit 1;
}

if ($autocap && ($Caption || $capfile)) {
    print "$0: specify only one of -caption or -capfile or -autocap\n";
    exit 1;
}

# sanity check
$stop =~ s|/$||; $stop =~ s|^/||;
# -reset forces -newkeys
$newkeys ||= $reset;

my $kw = {}; # keyword translation hash
sub mk_key_trans
{
    return if !$key_trans; # see end of function

    if (!open TRANSFILE, $key_trans) {
	print "$key_trans: $!\n";
	undef $key_trans;
	return;
    }
    while (<TRANSFILE>) {
	chomp; 
	s/\s*#.*//;	# strip comments
	next if /^$/;	# skip blank lines
	if (/(.*)=>(.*)/) { $k = $1; ($v = $2) ||= "" }
	else {
	    print "format error in $key_trans: '$_'\n";
	    next;
	}
	$k =~ s/(\s+)$//; # lose trailing blanks
	$k =~ s/(^\s+)//; # lose leading blanks
	$v =~ s/(\s+)$//; # lose trailing blanks
	$v =~ s/(^\s+)//; # lose leading blanks
	if ($k eq $v) {
	    print "$key_trans: key pairs ($k) are identical\n";
	    next;
	}
	$kw->{lc $k} = lc $v;
    }
    close TRANSFILE;
    undef $key_trans; # this function should only be run once
}

# Make a list of keywords from the argument.
#
# First, separate KEYWORDS using "/"'s and "-"'s. Thus,
#   /images/UnitedStates/California/SanFrancisco/GoldenGate/boy-n-dog-1.jpg
# turns into:
#   "united states", "california", "san francisco", "golden gate",
#    "boy" "n" "dog" "1.jpg"
#
# NOTE: splitting at - is a double-edged sword! "boy" and "dog" are two
# distinct keywords, whereas "world-trade-center" should be ONE keyword.
# This is a problem for filenames, not directory names, since directories
# have the convention of using caps to indentify words. Filenames should
# use _'s to separate words, but convention has allowed for -'s as well.
# Most of MY images don't use _'s but I'll have to start now! sigh.
#
# Because CapitalLettersSeparateWords, "SanFrancisco" => "san francisco"
# Thus, it is not advised to use ALLCAPS because they will be eliminated by:
# ALLCAPS => a l l c a p s   and all single-letter words will be removed.
#
# Now we are dealing with individual keywords at this point.
# Modify each according to these rules:
# 0) don't modify anything in double quotes. just skip it
#	useful for email addresses and URLs that have .'s and @'s and stuff
#
# 1) Lose file extensions:
#    "boy_n_dog_1.jpg" => "boy_n_dog_1"
#
# 2) Non-letters converted to spaces:
#	cobre_cathedral.jpg => "cobre cathedral"
#	the_white_house.jpg => "the white house"
# 	"July4" => "july"
#	"boys2men" => "boys men",
#	"joe's mom!" => "joe s mom"
#
# 3) dump all single letters:  "joe's mom!" => "joe" "mom"
#
# 4) leading and trailing whitespace is removed, and "padded" whitespace is
# condensed to one.
#
# 5) Keyword translations are applied according to the table defined
# by the file in $key_trans. The format of that file is:
#	expression1 => expression2
# one per line. For example,
#	bw => black and white
#	woman => women
#	girl => woman
#	female => girl
#	the =>
# Note that "the" converts to nothing, so the word is deleted.
# References unravel till there are no more keys to dereference.
# So, "female" => "women" (through three translations)
#
# Last, all two-letter words are deleted.
#
sub mk_keywords_from_path
{
    my $path = shift;
    my @a = ();
    my ($k, $v);

    return @a if !$path;

    # print "making keywords from \"$path\"\n";
    foreach my $s (split /[\/\-]/, $path) {
	# print "testing '$s'...";
	# save stuff in double quotes to protext from algorithm
	if ($s =~ s/"//g) {
	    $s = lc $s;
	} else {
	    $s =~ s/([A-Z])/ \l$1/g; # caps separate words (convert to lowercase)
	    $s =~ s/(big)?\.jpg$//; # lose trailing .'s and their extensions
	    $s =~ s/[^a-z]/ /g; # non-letters to spaces
	    $s =~ s/\b[a-z]\b//g; # lose all single letters
	}
	$s =~ s/(^\s+)//; # lose leading blanks
	$s =~ s/(\s+)$//; # lose trailing blanks
	$s =~ s/(\s\s+)/ /; # compress multiple spaces
	next if !$s;
	my $tried = {};
	for ($orig = $s; exists $kw->{$s} && $s; $s = $kw->{$s}) {
	    if (exists($tried->{$s})) {
		# girl => woman => women => girls => girl
		print "key translation loop: $orig";
		for ($s = $orig; $tried->{$s}; $orig = $s) {
		    print " => $tried->{$s}";
		    $s = $tried->{$s};
		    undef $tried->{$orig};
		}
		print "...\n";
		$s = ""; # don't add to list
		last;
	    }
	    $tried->{$s} = $kw->{$s};
	}

	next if (!$s || $s =~ /^..$/);

	# if defined, add to the list
	push @a, $s;
	# print "adding '$s'\n";
    }
    # print "\n";
    # sort and remove duplicates
    @a = sort @a;
    for (my $t = $#a; $t > 0; $t--) {
	# print qq(comparing "$a[$t]" with "$a[$t-1]"\n);
	if (!$a[$t] || $a[$t] eq $a[$t-1]) {
	    # print "spliced \"$a[$t]\"\n";
	    splice @a, $t, 1;
	}
	splice @a, $t-1, 1 if (!$a[$t-1]);
    }

    return @a;
}

# open the file (parameter), separate the keywords from the , or ; and
# call mk_keywords_from_path() on each one.
sub read_keyfile
{
    my $file = shift;
    my @keys;
    return 0 if !open FILE, $file;
    while (<FILE>) {
	chomp;
	s/#.*$//;
	next if /^$/;
	map { push @keys, mk_keywords_from_path($_) } split /[,;]/;
    }
    close FILE;
    return @keys;
}

sub print_report
{
    ($file, $info) = @_;

    #initialize @k to the item's keywords
    my @k = @{$info->Keywords()};

    # see if we need to prune out the toplevel keywords
    if (!$noprune && keys %$topkeys > 0) {
	# get rid of keywords listed in %topkeys hash.
	for (my $t = $#k; $t >= 0; $t--) {
	    splice @k, $t, 1 if exists $topkeys->{$k[$t]};
	}
    }

    # if user gave -terse, override $report_fmt
    $report_fmt = $terse_fmt if $terse;

    if (!$report_fmt) {
	print "path: $file\n";
	foreach (@iptc_headers) {
	    print qq($_: $a\n) if $a = $info->Attribute($_);
	}
	print "keywords: ", join "; ", @k;
	print "\n\n";
	return;
    }

    my $str;
    my $state = ($info->Attribute('province/state') || "");
    $state .= ", " if $state;
    my $city = ($info->Attribute('city') || "");
    $city .= ", " if $city;
    my $country = ($info->Attribute('country/primary location name') || "");
    my $location = "$city$state$country";
    foreach ($str = $report_fmt) {
	s/\\t/\t/g;
	s/\\n/\n/g;
	s/_path_/$file/ig;
	s/_fullpath_/cwd() . "\/$file"/eig;
	s/_state_/$state/ig;
	s/_country_/$country/ieg;
	s/_location_/$location/ig;
	s/_keywords_/join "$keyword_sep", @k/eig;
    }
    while ($str =~ /_(.*?)_/cg) {
	$s = ($info->Attribute($1) || "");
	$str =~ s/_($1?)_/$s/ig;
    }
    print "$str\n";
}

sub do_it
{
    my $img = shift;
    my $byline = $Byline;
    my $credit = $Credit;
    my $copyright = $Copyright;
    my $city = $City;
    my $state = $State;
    my $country = $Country;
    my $caption = $Caption;
    my $info;

    $img =~ m,(.*)/(.*),;
    my $path = $1;
    my $file = $2;
    if (!$path) {
	$path = cwd();
	$file = $img;
    }

    # Create new info object
    $info = create Image::IPTCInfo($img);
    if (! $info ) {
	print "$img: ", Image::IPTCInfo->Error, "\n";
	return;
    }

    # Get specific attributes...
    if (!$reset) {
	$byline ||= $info->Attribute('by-line');
	$credit ||= $info->Attribute('credit');
	$copyright ||= $info->Attribute('copyright notice');
	$city ||= $info->Attribute('city');
	$state ||= $info->Attribute('province/state');
	$country ||= $info->Attribute('country/primary location name');
	$caption ||= $info->Attribute('caption/abstract');
    }

    if ($autocap) {
	($capfile = "$path/$file") =~ s/(-sm|-big)?\.jpg$/.txt/;
	$caption = "";
	$capfile = "" if (! -T $capfile);
    }
    if ($capfile) {
	if (open FILE, $capfile) {
	    my $x = $/; undef $/;
	    $caption = <FILE>; # swallow whole file as a string
	    close FILE;
	    $/ = $x;
	    print "caption loaded from ", basename($capfile), "\n" if $verbose;
	    # avoid re-reading every time we loop. ($autocap resets anyway)
	    $capfile = "";
	} else {
	    print "$capfile: $!\n";
	}
    }

    if ($keyfile && $lastpath ne $path) {
	# Crawl up directory tree, looking for "Keywords" files. These are
	# to apply to all pages created at this level and up.
	# Only do this once per $path, so save/check $lastpath
	@masterkeys = (); # reset for each new directory!
	my $curdir = "$path/$keyfile";
	while ($curdir = dirname($curdir)) {
	    print "Reading Keyfile '$curdir/$keyfile': " if $verbose;
	    @a = read_keyfile("$curdir/$keyfile");
	    if ($#a >= 0) {
		print $#a + 1, " keywords\n" if $verbose;
		# if we're at the toplevel, sort and unique-ify, set %topkeys
		if ($curdir =~ /(^|\/)$stop$/o) {
		    # We have to do this anyway for @topkeys, so do it now...
		    @a = sort {$b cmp $a} @a;
		    for (my $t = $#a; $t >= 0; $t--) {
			if ($t && $a[$t] eq $a[$t-1] || !$a[$t]) {
			    splice @a, $t, 1;
			}
		    }
		    keys %$topkeys = $#a + 1; # pre-allocate hash size
		    map { $topkeys->{$_} = 1 } @a;
		}
		push @masterkeys, @a;
	    }
	    last if ($stop && ($curdir =~ /(^|\/)$stop$/o || $curdir eq "/"));
	}
	$lastpath = $path;
    }

    @keywords = @masterkeys;
    push @keywords, @arg_keywords;

    # save existing image's existing keywords if !$newkeys
    map { push @keywords, lc $_ } @{$info->Keywords()} if !$newkeys;

    $info->ClearKeywords();

    if ($autokeys) {
	# initialize keywords
	push @keywords, lc $country if $country;
	push @keywords, lc $state if $state;
	push @keywords, lc $city if $city;

	my @w = mk_keywords_from_path("$path/$file");
	push @keywords, @w;

	# sort, remove duplicates
	@keywords = sort @keywords;
	for (my $t = $#keywords; $t >= 0; $t--) {
	    if ($t && $keywords[$t] eq $keywords[$t-1] || !$keywords[$t]) {
		splice @keywords, $t, 1;
	    } 
	}
    }

    if (keys(%$rm_words) >= 0) {
	for (my $t = $#keywords; $t >= 0; $t--) {
	    if (exists $rm_words->{$keywords[$t]}) {
		print "Removing keyword '$keywords[$t]'\n" if $verbose; 
		splice @keywords, $t, 1;
	    }
	}
    }

    map { $info->AddKeyword( $_ ) if $_ } @keywords;

    $file = $img if !$file;

    map { $info->SetAttribute($_, "") } @iptc_headers if $reset;
    $info->SetAttribute('object name', $file);
    $info->SetAttribute('by-line', $byline);
    $info->SetAttribute('credit', $credit);
    $info->SetAttribute('country/primary location name', $country);
    $info->SetAttribute('province/state', $state);
    $info->SetAttribute('city', $city);
    $info->SetAttribute('copyright notice', $copyright);
    $info->SetAttribute('caption/abstract', $caption);

    print_report("$img", $info);
    if ($save) {
	if ($dir) {
	    $info->SaveAs("$dir/$file") ;
	    print "saved image to $dir/$file...\n";
	} else {
	    $info->Save();
	    print "updated $img...\n";
	}
    }
}

sub dofile
{
    do_it ($_) if (/\.(jpg|gif)$/i);
}

if ($arg_keys) {
    print "New keywords: ";
    foreach (split /,/, $arg_keys) {
	s/^\s*(.*)\s*$/$1/;
	if (s/^-//) { $rm_words->{lc $_} = 1; print "(removing $_) " }
	else { push @arg_keywords, lc $_; print lc $_, " " }
    }
    print "\n";
}

# see if this is a fullpath, if so do ONCE per loop! (no tree traverse)
if ($keyfile =~ m|^/|) {
    @masterkeys = read_keyfile($keyfile);
    $keyfile = "";
}

mk_key_trans();

while (my $arg = shift) {
    if (-d $arg) {
	$arg =~ s,/$,,;
	if ($recurse) {
	    find({ wanted => \&dofile, no_chdir => 1 }, $arg);
	} else {
	    while (<$arg/*>) { dofile($_) }
	}
    } else {
	do_it ($arg);
    }
}
